@@ -5,6 +5,8 @@ import time
 from pydantic import BaseModel, PrivateAttr
 from fastapi.encoders import jsonable_encoder
 from .gql import GQLException
+from .execute import gql
+from devtools import debug
 
 
 def d_rez(refresh: bool = False, use_stale: bool = False, returns: T.Type = None):
@@ -121,6 +123,8 @@ NodeModel = T.TypeVar("NodeModel", bound="Node")
 ResolverType = T.TypeVar("ResolverType", bound="Resolver")
 ResolverTempType = T.TypeVar("ResolverTempType", bound="Resolver")
 
+GQL_D = T.Dict[str, T.Any]
+
 
 class Node(BaseModel):
     _cache: CacheManager = PrivateAttr(default_factory=CacheManager)
@@ -130,6 +134,14 @@ class Node(BaseModel):
 
     id: str
 
+    class Config:
+        validate_assignment = True
+
+    class Dgraph:
+        typename: T.ClassVar[str]
+        resolver: T.ClassVar[T.Type[ResolverType]]
+        payload_node_name: str
+
     # TODO make equality and hashing
 
     @property
@@ -217,38 +229,134 @@ class Node(BaseModel):
             self.cache.replace(key=name, cache=obj.cache.get(name))
         return self.cache.get_val(name)
 
-    class Dgraph:
-        typename: T.ClassVar[str]
-        resolver: T.ClassVar[T.Type[ResolverType]]
-
     @classmethod
     @property
     def resolver(cls) -> ResolverType:
         return cls.Dgraph.resolver()
 
+    def hydrate(self, new_node: NodeModel) -> None:
+        """Turns this node into the new node"""
+        for field_name in new_node.__fields__.keys():
+            new_field = getattr(new_node, field_name)
+            old_field = getattr(self, field_name)
+            if new_field != old_field:
+                setattr(self, field_name, getattr(new_node, field_name))
+        for private_attr_name in new_node.__private_attributes__.keys():
+            setattr(self, private_attr_name, getattr(new_node, private_attr_name))
+
     """ADD REFRESH BEFORE ADDING ADD"""
 
-    async def refresh(self: NodeModel) -> NodeModel:
-        """ In future, just replace all fields including cache w new ones... """
-        return await self._used_resolver._get({"id": self.id})
+    async def refresh(self: NodeModel) -> None:
+        new_node = await self._used_resolver._get({"id": self.id})
+        self.hydrate(new_node=new_node)
 
-    def add_old(self: NodeModel, upsert: bool = False) -> NodeModel:
-        # TODO
-        """Include options to add relationships -> like taught_by: {"username": "malia"}... autogen this!"""
-        """Must include a refresh option too, no?"""
-        input = self.dict(exclude={"id"})  # actually do INPUT type when this is ready
-        # return exact obj w current resolver configs
+    """CRUDS"""
 
-    def update(self: NodeModel) -> NodeModel:
-        # TODO
-        pass
+    def base_set_remove_d(self) -> T.Tuple[GQL_D, GQL_D]:
+        set_d: GQL_D = {}
+        remove_d: GQL_D = {}
 
-    def delete(self) -> bool:
-        # TODO
-        pass
+        curr_dict = self.dict()
+        for field_name, val in self._original_dict.items():
+            if field_name not in curr_dict:
+                # it was removed
+                remove_d[field_name] = val
+            elif val != curr_dict[field_name]:
+                # it was changed
+                set_d[field_name] = curr_dict[field_name]
+        # now set all the new ones
+        new_field_names = set(curr_dict.keys()) - set(self._original_dict.keys())
+        for new_field_name in new_field_names:
+            set_d[new_field_name] = curr_dict[new_field_name]
 
-    class Config:
-        validate_assignment = True
+        return set_d, remove_d
+
+    async def update(self, given_resolver: ResolverType = None) -> bool:
+        """Will usually be overwritten assuming there are relationshops with this node"""
+        set_d, remove_d = self.base_set_remove_d()
+        return await self._update(
+            set_d=set_d, remove_d=remove_d, given_resolver=given_resolver
+        )
+
+    @staticmethod
+    def relationships_set_remove_d(
+        to_set: T.Dict[str, T.Optional[T.Union[NodeModel, T.List[NodeModel]]]],
+        to_remove: T.Dict[str, T.Optional[T.Union[NodeModel, T.List[NodeModel]]]],
+    ) -> T.Tuple[GQL_D, GQL_D]:
+        set_d: GQL_D = {}
+        for field_name, node_or_nodes in to_set.items():
+            if node_or_nodes is None:
+                continue
+            if type(node_or_nodes) is list:
+                val = {"id": n.id for n in node_or_nodes}
+            else:
+                node_or_nodes: NodeModel
+                val = {"id": node_or_nodes.id}
+            set_d[field_name] = val
+        remove_d: GQL_D = {}
+        for field_name, node_or_nodes in to_remove.items():
+            if node_or_nodes is None:
+                continue
+            if type(node_or_nodes) is list:
+                val = {"id": n.id for n in node_or_nodes}
+            else:
+                node_or_nodes: NodeModel
+                val = {"id": node_or_nodes.id}
+            remove_d[field_name] = val
+        return set_d, remove_d
+
+    async def _update(
+        self,
+        set_d: GQL_D,
+        remove_d: GQL_D,
+        given_resolver: T.Optional[ResolverType] = None,
+        print_update_d: bool = False,
+    ) -> bool:
+
+        if not set_d and not remove_d:
+            print("NOTHING TO UPDATE!")
+            return False
+
+        variables = {"set": set_d, "remove": remove_d, "filter": {"id": self.id}}
+
+        if print_update_d:
+            debug(variables)
+
+        resolver = given_resolver or self._used_resolver
+        query_str = resolver.make_update_mutation_str()
+        j = await gql(query_str=query_str, variables=variables)
+        node_d_lst = j["data"][f"update{self.Dgraph.typename}"][
+            self.Dgraph.payload_node_name
+        ]
+
+        if not node_d_lst:
+            # raise Exception("No update was registered!!")
+            print("NO UPDATE WAS REGISTERED")
+            return False
+
+        node_d = node_d_lst[0]
+        new_node = resolver.parse_obj_nested(node_d)
+        self.hydrate(new_node=new_node)
+
+        return True
+
+    async def delete(self, given_resolver: ResolverType = None) -> bool:
+        resolver = given_resolver or self._used_resolver
+        query_str = resolver.make_delete_mutation_str()
+        variables = {"filter": {"id": self.id}}
+        j = await gql(query_str=query_str, variables=variables)
+        node_d_lst = j["data"][f"delete{self.Dgraph.typename}"][
+            self.Dgraph.payload_node_name
+        ]
+        if not node_d_lst:
+            # raise Exception("No update was registered!!")
+            print("NO DELETE WAS REGISTERED")
+            return False
+        node_d = node_d_lst[0]
+        node = resolver.parse_obj_nested(node_d)
+        node._deleted = True
+        self.hydrate(new_node=node)
+        return True
 
 
 from .resolver import Resolver
